#!/usr/bin/python

import tornado.web

from . import auth
from . import base
from . import ui_modules

class CreateHandler(auth.CacheMixin, base.BaseHandler):
	MODES = ("paste", "upload")

	def get(self):
		mode = self.get_argument("mode", "paste")
		if not mode in self.MODES:
			raise tornado.web.HTTPError(400)

		self.render("nopaste/create.html", mode=mode,
			max_size=self._max_size)

	@base.ratelimit(minutes=15, requests=5)
	def post(self):
		mode = self.get_argument("mode")
		if not mode in self.MODES:
			raise tornado.web.HTTPError(400)

		mimetype = "text/plain"

		if mode == "paste":
			subject = self.get_argument("subject", None)
			content = self.get_argument("content", "").encode("utf-8")

		elif mode == "upload":
			for f in self.request.files.get("file"):
				subject  = f.filename
				content  = f.body
				mimetype = f.content_type
				break

		# Check maximum size
		if len(content) > self._max_size:
			raise tornado.web.HTTPError(400,
				"You cannot upload more than %s bytes" % self._max_size)

		expires = self.get_argument("expires", "0")
		try:
			expires = int(expires)
		except (TypeError, ValueError):
			expires = None

		uid = self.backend.nopaste.create(subject, content, mimetype=mimetype,
			expires=expires, account=self.current_user, address=self.get_remote_ip())

		if uid:
			return self.redirect("/view/%s" % uid)

		raise tornado.web.HTTPError(500)

	@property
	def _max_size(self):
		# Authenticated users are allowed to upload up to 25MB
		if self.current_user:
			return 25 * (1024 ** 2)

		# The rest is only allowed to upload up to 2MB
		return 2 * (1024 ** 2)


class RawHandler(base.BaseHandler):
	def get(self, uid):
		entry = self.backend.nopaste.get(uid)
		if not entry:
			raise tornado.web.HTTPError(404)

		# Set the filename
		self.set_header("Content-Disposition", "inline; filename=\"%s\"" % entry.subject)

		# Set mimetype
		self.set_header("Content-Type", entry.mimetype)

		# Set expiry headers
		self.set_expires(3600)

		# Send content
		content = self.backend.nopaste.get_content(entry.uuid)
		self.finish(content)


class ViewHandler(auth.CacheMixin, base.BaseHandler):
	def get(self, uid):
		entry = self.backend.nopaste.get(uid)
		if not entry:
			raise tornado.web.HTTPError(404)

		# Fetch the content if the output should be displayed
		if entry.mimetype.startswith("text/"):
			content = self.backend.nopaste.get_content(entry.uuid)
		else:
			content = None

		# Set expiry headers
		self.set_expires(3600)

		self.render("nopaste/view.html", entry=entry, content=content)


class CodeModule(ui_modules.UIModule):
	def render(self, content):
		return self.render_string("nopaste/modules/code.html", content=content)

	def javascript_files(self):
		return "js/prettify.js"

	def css_files(self):
		return "css/prettify.css"

	def embedded_javascript(self):
		return """
			// Run pretty print
			prettyPrint();
		"""
